#include "PageLayout.hh"


#include <iostream>
using namespace std;

PageLayout::PageLayout()
  : Glib::ObjectBase(typeid(PageLayout)),
    prop_row_spacing (*this, "row-spacing", 10)
{
  /*
  set_has_window(false);
  set_redraw_on_allocate(false);
  */
}

/*
void PageLayout::on_add(Gtk::Widget* widget)
{
  widget->set_parent(*this);
}
*/

bool PageLayout::on_draw(const Cairo::RefPtr<Cairo::Context>& cr)
{
  Gtk::Allocation alloc = get_allocation();
  
  bool x = Gtk::Layout::on_draw(cr);
  
  cr->set_source_rgba(0, 0.8, 0, 1);
  cr->move_to(0, 0);
  cr->line_to(alloc.get_width(), alloc.get_height());
  cr->move_to(alloc.get_width(), 0);
  cr->line_to(0, alloc.get_height());
  cr->stroke();
  
  return x;
}

void PageLayout::get_preferred_width_vfunc(int& minimum_width, int& natural_width) const
{
  minimum_width = natural_width = 0;
  
  for(const Gtk::Widget* widget : get_children())
  {
    int widget_minimum_width, widget_natural_width;
    
    widget->get_preferred_width(widget_minimum_width, widget_natural_width);
    
    minimum_width = std::max(minimum_width, widget_minimum_width);
    natural_width = std::max(natural_width, widget_natural_width);
  }
  
  cout << "PageLayout::get_preferred_width_vfunc" << endl;
}

void PageLayout::get_preferred_height_vfunc(int& minimum_height, int& natural_height) const
{
  minimum_height = natural_height = 0;
  int n_vis_children = 0;
  
  for(const Gtk::Widget* widget : get_children())
  {
    if(!widget->get_visible())
      continue;
    
    ++n_vis_children;
    
    int widget_minimum_height, widget_natural_height;
    
    widget->get_preferred_height(widget_minimum_height, widget_natural_height);
    
    minimum_height += widget_minimum_height;
    natural_height += widget_natural_height;
  }
  
  if(n_vis_children > 1)
  {
    int margin = (n_vis_children - 1) * prop_row_spacing.get_value();
    minimum_height += margin;
    natural_height += margin;
  }
  
  cout << "PageLayout::get_preferred_height_vfunc" << endl;
}

void PageLayout::on_size_allocate(Gtk::Allocation& allocation)
{
  cout << "PageLayout::on_size_allocate" << endl;
  
  set_allocation(allocation);
  
  if(get_realized() && get_has_window())
  {
    std::cout << "move/resize" << endl;
    
    get_window()->move_resize(allocation.get_x(),
                              allocation.get_y(),
                              allocation.get_width(),
                              allocation.get_height());
  }
  
  int x = allocation.get_x(),
      y = allocation.get_y();
  
  int n_expanding = 0,
      min_height = 0;
  
  for(Gtk::Widget* widget : get_children())
  {
    if(widget->get_parent() != this)
    {
      std::cerr << "Widget has invalid parent" << endl;
    }
    
    if(!widget->get_visible())
      continue;
    
    int widget_minimum_height, widget_natural_height;
    
    widget->get_preferred_height(widget_minimum_height, widget_natural_height);
    
    min_height += widget_minimum_height;
    
    if(widget->get_vexpand())
      ++n_expanding;
  }
  
  int expanding_height = 0;
  
  if(n_expanding > 0)
    expanding_height = (allocation.get_height() - min_height) / (n_expanding);
  
  for(Gtk::Widget* widget : get_children())
  {
    if(!widget->get_visible())
      continue;
    
    int widget_minimum_width, widget_natural_width;
    
    widget->get_preferred_width(widget_minimum_width, widget_natural_width);
    
    const int available_width = allocation.get_width();
    int width = 0;
    
    if(widget->get_hexpand())
    {
      width = available_width;
    }
    else
    {
      width = widget_minimum_width;
    }
    
    Gtk::Allocation widget_alloc;
    
    widget_alloc.set_width(width);
    widget_alloc.set_x(x + (available_width - width)/2);
    widget_alloc.set_y(y);
    
    int widget_minimum_height, widget_natural_height;
    
    widget->get_preferred_height(widget_minimum_height, widget_natural_height);
    
    if(widget->get_vexpand())
    {
      widget_alloc.set_height(widget_minimum_height + expanding_height);
    }
    else
    {
      widget_alloc.set_height(widget_minimum_height);
    }
    
    if(widget->get_realized() && widget->get_has_window())
    {
      std::cout << "Before allocation: " << std::endl;
      
      dump_cairo_region(widget->get_window()->get_clip_region());
    }
    
    widget->size_allocate(widget_alloc);
    
    if(widget->get_realized() && widget->get_has_window())
    {
      std::cout << "After allocation: " << std::endl;
      
      dump_cairo_region(widget->get_window()->get_clip_region());
    }
      
    /*
    if(get_realized() && get_has_window())
    {
      std::cout << "Own clip region:" << endl;
      
      dump_cairo_region(get_window()->get_clip_region());
    }
    */
    
    y += prop_row_spacing.get_value() + widget_alloc.get_height();
  }
}

void PageLayout::dump_cairo_region(Cairo::RefPtr<Cairo::Region> region)
{
  for(int i = 0; i < region->get_num_rectangles(); ++i)
  {
    auto rect = region->get_rectangle(i);
    
    std::cout << "Rectangle x = " << rect.x
              << ", y = " << rect.y
              << " | width = " << rect.width
              << ", height = " << rect.height << std::endl;
  }
}
